# -*- coding: utf-8 -*-
# Zilliqa Python Library
# Copyright (C) 2019  Gully Chen
# MIT License
"""
pyzil.crypto.drbg
~~~~~~~~~~~~

HMAC DRBG implementation of Nist SP 800-90A

:copyright: (c) 2019 by Gully Chen.
:license: MIT License, see LICENSE for more details.
"""

import hmac
import logging
import secrets
import hashlib


ENTROPY_BYTES = 32


def randbelow_drbg(boundary, nonce=None):
    """Return a random int in the range (0, n)."""
    if boundary <= 0:
        raise ValueError("boundary cannot be less than zero")
    num_bytes = (boundary.bit_length() + 7) // 8

    if nonce is None:
        nonce = secrets.token_bytes(ENTROPY_BYTES)

    entropy = nonce + secrets.token_bytes(ENTROPY_BYTES)

    drbg = HMAC_DRBG(entropy, personalization_string=b"pyzil_hmac_drbg")
    while True:
        rand_bytes = drbg.generate(num_bytes)
        if rand_bytes is None:
            logging.critical("reseed HMAC DRBG")
            drbg.reseed(nonce + secrets.token_bytes(ENTROPY_BYTES))
            continue

        r = int.from_bytes(rand_bytes, byteorder="big")
        if r < boundary:
            break
    return r


class HMAC_DRBG(object):
    """
    A Python implementation of HMAC_DRBG, as specified by NIST SP 800-90A.
    https://github.com/fpgaminer/python-hmac-drbg/blob/master/hmac_drbg/hmac_drbg.py
    """
    def __init__(self, entropy, requested_security_strength=256, personalization_string=b""):
        if requested_security_strength > 256:
            raise RuntimeError("requested_security_strength cannot exceed 256 bits.")

        # Modified from Appendix D, which specified 160 bits here
        if len(personalization_string) * 8 > 256:
            raise RuntimeError("personalization_string cannot exceed 256 bits.")

        if requested_security_strength <= 112:
            self.security_strength = 112
        elif requested_security_strength <= 128:
            self.security_strength = 128
        elif requested_security_strength <= 192:
            self.security_strength = 192
        else:
            self.security_strength = 256

        if (len(entropy) * 8 * 2) < (3 * self.security_strength):
            raise RuntimeError("entropy must be at least %f bits." % (1.5 * self.security_strength))

        self.K = None
        self.V = None
        self.reseed_counter = 1
        self._instantiate(entropy, personalization_string)

    def _hmac(self, key, data):
        return hmac.new(key, data, hashlib.sha256).digest()

    def _update(self, provided_data=None):
        self.K = self._hmac(self.K, self.V + b"\x00" + (b"" if provided_data is None else provided_data))
        self.V = self._hmac(self.K, self.V)

        if provided_data is not None:
            self.K = self._hmac(self.K, self.V + b"\x01" + provided_data)
            self.V = self._hmac(self.K, self.V)

    def _instantiate(self, entropy, personalization_string):
        seed_material = entropy + personalization_string

        self.K = b"\x00" * 32
        self.V = b"\x01" * 32

        self._update(seed_material)
        self.reseed_counter = 1

    def reseed(self, entropy):
        if (len(entropy) * 8) < self.security_strength:
            raise RuntimeError("entropy must be at least %f bits." % self.security_strength)

        self._update(entropy)
        self.reseed_counter = 1

    def generate(self, num_bytes, requested_security_strength=256):
        if (num_bytes * 8) > 7500:
            raise RuntimeError("generate cannot generate more than 7500 bits in a single call.")

        if requested_security_strength > self.security_strength:
            raise RuntimeError(
                "requested_security_strength exceeds this instance's security_strength (%d)" % self.security_strength)

        if self.reseed_counter >= 10000:
            return None

        temp = b""

        while len(temp) < num_bytes:
            self.V = self._hmac(self.K, self.V)
            temp += self.V

        self._update(None)
        self.reseed_counter += 1

        return temp[:num_bytes]



# -*- coding: utf-8 -*-
# Zilliqa Python Library
# Copyright (C) 2019  Gully Chen
# MIT License
"""
pyzil.crypto.schnorr
~~~~~~~~~~~~

Zilliqa schnorr signature

:copyright: (c) 2019 by Gully Chen.
:license: MIT License, see LICENSE for more details.
"""

import secrets
import hashlib
from typing import Optional

from fastecdsa import keys
from fastecdsa import point
from fastecdsa import curve

from pyzil.crypto.drbg import randbelow_drbg


CURVE = curve.secp256k1
CURVE_BITS = 256
ENCODED_SIZE = CURVE_BITS // 8

SECP256K1_TAG_PUBKEY_EVEN = b"\x02"
SECP256K1_TAG_PUBKEY_ODD = b"\x03"
SECP256K1_TAG_PUBKEY_UNCOMPRESSED = b"\x04"


def gen_private_key() -> int:
    """Generate a private key."""
    return keys.gen_private_key(CURVE)


def get_public_key(private_key: int) -> point.Point:
    """Get public key from a private key."""
    return keys.get_public_key(private_key, CURVE)


def encode_signature(r: int, s: int, size=ENCODED_SIZE) -> bytes:
    """encode signature to bytes."""
    r = r.to_bytes(size, "big")
    s = s.to_bytes(size, "big")
    return r + s


def decode_signature(signature: bytes) -> (int, int):
    """decode signature bytes to integer (r, s)."""
    size = len(signature) // 2
    r = int.from_bytes(signature[0:size], "big")
    s = int.from_bytes(signature[size:], "big")
    return r, s


def encode_public(x: int, y: int, compressed=True) -> bytes:
    """encode public key to bytes."""
    if compressed:
        tag = SECP256K1_TAG_PUBKEY_ODD if (y & 0x01) else SECP256K1_TAG_PUBKEY_EVEN
        return tag + x.to_bytes(ENCODED_SIZE, "big")
    else:
        tag = SECP256K1_TAG_PUBKEY_UNCOMPRESSED
        return tag + x.to_bytes(ENCODED_SIZE, "big") + y.to_bytes(ENCODED_SIZE, "big")


def decode_public(pub_key: bytes) -> point.Point:
    """decode public key from bytes to Point"""
    try:
        if len(pub_key) == 33:
            # compressed format
            y_odd = pub_key[0:1] == SECP256K1_TAG_PUBKEY_ODD
            x = int.from_bytes(pub_key[1:], "big")

            y_y = pow(x, 3) + CURVE.a * x + CURVE.b
            y = mod_sqrt(y_y, CURVE.p, y_odd)

        elif len(pub_key) == 65:
            tag = pub_key[0:1]
            assert tag == SECP256K1_TAG_PUBKEY_UNCOMPRESSED

            x = int.from_bytes(pub_key[1:ENCODED_SIZE + 1], "big")
            y = int.from_bytes(pub_key[ENCODED_SIZE + 1:], "big")

        else:
            raise NotImplementedError

        return point.Point(x, y, CURVE)
    except:
        raise ValueError("The public key could not be parsed or is invalid")


def mod_sqrt(n: int, p: int, is_odd: bool) -> int:
    """ Find Square Root under Modulo p
    Given a number 'n' and a prime 'p', find square root of n under modulo p if it exists.
    https://www.geeksforgeeks.org/find-square-root-under-modulo-p-set-1-when-p-is-in-form-of-4i-3/
    """
    n %= p
    y = pow(n, (p + 1) // 4, p)

    y_odd = bool(y & 0x01)
    if y_odd != is_odd:
        y = p - y

    assert (y * y) % p == n
    return y


def sign(bytes_msg: bytes, bytes_private: bytes, retries=10) -> Optional[bytes]:
    """sign bytes message with private key."""
    for i in range(retries):
        # k = secrets.randbelow(CURVE.q)
        k = randbelow_drbg(CURVE.q, nonce=bytes_private + bytes_msg)
        if k == 0:
            continue
        signature = sign_with_k(bytes_msg, bytes_private, k)
        if signature:
            return signature
    return None


def sign_with_k(bytes_msg: bytes,
                bytes_private: bytes,
                k: int) -> Optional[bytes]:
    """Sign bytes using Zilliqa schnorr signature algorithm."""
    assert isinstance(bytes_msg, bytes)
    private_key = int.from_bytes(bytes_private, "big")

    order = CURVE.q

    Q = CURVE.G * k
    bytes_Q_x = encode_public(Q.x, Q.y)

    pub_key = keys.get_public_key(private_key, CURVE)
    bytes_pub_x = encode_public(pub_key.x, pub_key.y)

    hasher = hashlib.sha256()
    hasher.update(bytes_Q_x + bytes_pub_x + bytes_msg)
    r = hasher.digest()
    r = int.from_bytes(r, "big") % order
    s = (k - r * private_key) % order
    if r == 0 or s == 0:
        return None

    return encode_signature(r, s)


def verify(bytes_msg: bytes,
           signature: bytes,
           bytes_public: bytes) -> bool:
    """verify schnorr signature with public key."""
    assert isinstance(bytes_msg, bytes)
    assert isinstance(signature, bytes)
    pub_key = decode_public(bytes_public)

    r, s = decode_signature(signature)
    n, G = CURVE.q, CURVE.G

    if not s or s >= n:
        return False
    if not r or r >= pow(2, CURVE_BITS):
        return False

    sG = s * G
    rW = r * pub_key
    Q = sG + rW

    bytes_Q_x = encode_public(Q.x, Q.y)
    bytes_pub_x = encode_public(pub_key.x, pub_key.y)

    hasher = hashlib.sha256()
    hasher.update(bytes_Q_x + bytes_pub_x + bytes_msg)
    v = hasher.digest()
    v = int.from_bytes(v, "big")
    v = v % n

    return v == r



# -*- coding: utf-8 -*-
# Zilliqa Python Library
# Copyright (C) 2019  Gully Chen
# MIT License
"""
pyzil.crypto.zilkey
~~~~~~~~~~~~

Zilliqa Key

:copyright: (c) 2019 by Gully Chen.
:license: MIT License, see LICENSE for more details.
"""

import json
import uuid
from typing import Union, Optional
from collections import namedtuple

from pyzil.common import utils
from pyzil.crypto import tools, schnorr, bech32


# zilliqa address takes the last 20 bytes from hash digest of public key
ADDRESS_NUM_BYTES = 20
ADDRESS_STR_LENGTH = ADDRESS_NUM_BYTES * 2


def is_valid_address(address: str) -> bool:
    """Return True if address is valid."""
    if address.lower().startswith("0x"):
        address = address[2:]
    if len(address) != ADDRESS_STR_LENGTH:
        return False
    # noinspection PyBroadException
    try:
        utils.hex_str_to_int(address)
    except Exception:
        return False
    return True


def to_valid_address(address: str) -> Optional[str]:
    """Return lower case address if address is valid."""
    if is_bech32_address(address):
        address = from_bech32_address(address)

    if not is_valid_address(address):
        return None
    address = address.lower()
    if address.startswith("0x"):
        address = address[2:]
    return address


def to_checksum_address(address: str, prefix="0x") -> Optional[str]:
    """Convert address to checksum address."""
    if is_bech32_address(address):
        address = from_bech32_address(address)

    if not is_valid_address(address):
        return None

    address = address.lower().replace("0x", "")
    address_bytes = utils.hex_str_to_bytes(address)
    v = utils.bytes_to_int(tools.hash256_bytes(address_bytes))

    checksum_address = prefix
    for i, c in enumerate(address):
        if not c.isdigit():
            if v & (1 << 255 - 6 * i):
                c = c.upper()
            else:
                c = c.lower()
        checksum_address += c

    return checksum_address


def is_valid_checksum_address(address: str) -> bool:
    """Return True if address is valid checksum address."""
    if not is_valid_address(address):
        return False
    return to_checksum_address(address) == address


def to_bech32_address(address: str) -> Optional[str]:
    """Convert 20 bytes address to bech32 address."""
    if not is_valid_address(address):
        return None
    return bech32.encode("zil", utils.hex_str_to_bytes(address))


def from_bech32_address(bech32_address: str) -> Optional[str]:
    """Convert bech32 address to 20 bytes address."""
    data = bech32.decode("zil", bech32_address)
    if data is None:
        return None
    address = utils.bytes_to_hex_str(bytes(data))
    return to_valid_address(address)


def is_bech32_address(bech32_address: str) -> bool:
    """Return True if address is valid bech32 address."""
    if not bech32_address.startswith("zil1"):
        return False
    return from_bech32_address(bech32_address) is not None


def normalise_address(address: str) -> str:
    """Check address format, return checksum address."""
    if is_bech32_address(address):
        address = from_bech32_address(address)
        if not address:
            raise ValueError("Invalid address format")
        return to_checksum_address(address)

    if not is_valid_checksum_address(address):
        raise ValueError("Invalid address format")
    return address


KeyPair = namedtuple("KeyPair", ["public", "private"])


class ZilKey:
    """ Zilliqa Key """
    def __init__(self, public_key=None, private_key=None):
        assert public_key or private_key, "public or private key is required"
        if isinstance(public_key, str):
            public_key = utils.hex_str_to_bytes(public_key)
        if isinstance(private_key, str):
            private_key = utils.hex_str_to_bytes(private_key)

        self._bytes_public = public_key
        self._bytes_private = private_key

        # the _pub_key is a Point on curve
        self._public_key = None
        # the private_key is big integer less than curve order
        self._private_key = None

        self._generate_keys()

    def _generate_keys(self):
        if self._bytes_private:
            self._private_key = utils.bytes_to_int(self._bytes_private)
            assert self._private_key < schnorr.CURVE.q
        if self._bytes_public:
            self._public_key = schnorr.decode_public(self._bytes_public)

        if self._private_key and self._public_key:
            _pub_key = schnorr.get_public_key(self._private_key)
            assert _pub_key == self._public_key, "public/private key mismatch"

        # generate public key from private key
        if self._private_key and not self._public_key:
            self._public_key = schnorr.get_public_key(self._private_key)

    @property
    def encoded_public_key(self):
        """bytes of public key."""
        return schnorr.encode_public(self._public_key.x, self._public_key.y)

    @property
    def encoded_private_key(self):
        """bytes of private key."""
        return self._private_key and utils.int_to_bytes(self._private_key)

    @property
    def keypair_bytes(self) -> KeyPair:
        """bytes of key pair."""
        return KeyPair(self.encoded_public_key, self.encoded_private_key)

    @property
    def keypair_str(self) -> KeyPair:
        """hex string of key pair."""
        str_pub = utils.bytes_to_hex_str(self.encoded_public_key)
        str_private = self._private_key and utils.int_to_hex_str(self._private_key)
        return KeyPair(str_pub, str_private)

    @property
    def address(self) -> str:
        addr_bytes = tools.hash256_bytes(self.keypair_bytes.public)
        return utils.bytes_to_hex_str(addr_bytes)[-ADDRESS_STR_LENGTH:]

    @property
    def checksum_address(self) -> str:
        return to_checksum_address(self.address)

    @property
    def bech32_address(self) -> str:
        return to_bech32_address(self.address)

    def __str__(self):
        return str(self.keypair_str)

    def __eq__(self, other):
        return self._public_key == other._public_key and self._private_key == other._private_key

    # Zilliqa schnorr signature
    def sign(self, message: bytes) -> bytes:
        """Sign bytes message with private key, return bytes"""
        if not self._private_key:
            raise RuntimeError("missing private key")

        message = utils.ensure_bytes(message)

        signature = schnorr.sign(message, self.keypair_bytes.private)
        if signature is None:
            raise RuntimeError("failed to sign")
        return signature

    def sign_str(self, message: str) -> str:
        """Sign bytes message with private key, return hex string"""
        message = utils.ensure_bytes(message)
        return utils.bytes_to_hex_str(self.sign(message))

    def verify(self, signature: Union[str, bytes], message: Union[str, bytes]) -> bool:
        """Verify signature with public key."""
        if isinstance(signature, str):
            signature = utils.hex_str_to_bytes(signature)

        message = utils.ensure_bytes(message)

        return schnorr.verify(message, signature, self.keypair_bytes.public)

    @classmethod
    def generate_new(cls):
        """Generate new zilliqa key"""
        zil_key = cls(private_key=utils.int_to_bytes(schnorr.gen_private_key()))
        return zil_key

    @classmethod
    def load_mykey_txt(cls, key_file="mykey.txt"):
        """Load Zilliqa key from mykey.txt."""
        with open(key_file, "r") as f:
            str_pub, str_private = f.read().split()
            return ZilKey(public_key=str_pub, private_key=str_private)

    def save_mykey_txt(self, key_file="mykey.txt"):
        """Save key to mykey.txt."""
        with open(key_file, "w") as f:
            str_pub, str_private = self.keypair_str
            f.write(str_pub.upper() + " " + str_private.upper())

    @classmethod
    def load_keystore(cls, password: str, keystore_file):
        """Load Zilliqa key from keystore json file."""
        if hasattr(keystore_file, "read"):
            keystore = json.load(keystore_file)
        else:
            with open(keystore_file) as f:
                keystore = json.load(f)

        checksum_address = to_checksum_address(keystore["address"])
        ciphertext = utils.hex_str_to_bytes(keystore["crypto"]["ciphertext"])
        iv = utils.hex_str_to_bytes(keystore["crypto"]["cipherparams"]["iv"])

        kdf_method = keystore["crypto"]["kdf"]
        kdfparams = keystore["crypto"]["kdfparams"]

        derived_key = tools.gen_derived_key(password, kdf_method, kdfparams)

        cipher = keystore["crypto"]["cipher"].encode()
        message = derived_key[16:32] + ciphertext + iv + cipher
        mac_generated = tools.hmac_hash256(derived_key, msg=message)

        mac_from_file = utils.hex_str_to_bytes(keystore["crypto"]["mac"])
        if not tools.compare_digest(mac_generated, mac_from_file):
            raise ValueError("Invalid password or keystore file")

        key = derived_key[:16]
        private_key = tools.aes_ctr_decrypt(key, iv, ciphertext)
        zilkey = cls(private_key=private_key)
        if not tools.compare_digest(checksum_address, zilkey.checksum_address):
            raise ValueError("Invalid keystore file, address mismatch")
        return zilkey

    def save_keystore(self, password: str, kdf_method: str="pbkdf2",
                      keystore_file=None) -> dict:
        """Save Zilliqa key to keystore, format details on
        https://github.com/ethereum/wiki/wiki/Web3-Secret-Storage-Definition.
        """
        address = self.address
        cipher = "aes-128-ctr"
        salt = utils.rand_bytes(32)
        iv = utils.rand_bytes(16)
        kdfparams = {
            "salt": utils.bytes_to_hex_str(salt),
            "n": 8192,
            "c": 262144,
            "r": 8,
            "p": 1,
            "dklen": 32
        }
        derived_key = tools.gen_derived_key(password, kdf_method, kdfparams)

        key = derived_key[:16]
        ciphertext = tools.aes_ctr_encrypt(key, iv,
                                           ciphertext=self.encoded_private_key)
        cipher_bytes = cipher.encode()
        message = derived_key[16:32] + ciphertext + iv + cipher_bytes
        mac_generated = tools.hmac_hash256(derived_key, msg=message)

        keystore = {
            "address": address,
            "crypto": {
                "cipher": cipher,
                "cipherparams": {
                    "iv": utils.bytes_to_hex_str(iv),
                },
                "ciphertext": utils.bytes_to_hex_str(ciphertext),
                "kdf": kdf_method,
                "kdfparams": kdfparams,
                "mac": utils.bytes_to_hex_str(mac_generated),
            },
            "id": str(uuid.uuid4()),
            "version": 3,
        }

        if keystore_file is None:
            return keystore
        if hasattr(keystore_file, "write"):
            json.dump(keystore, keystore_file)
        else:
            with open(keystore_file, "w") as f:
                json.dump(keystore, f)
        return keystore



# -*- coding: utf-8 -*-
# Zilliqa Python Library
# Copyright (C) 2019  Gully Chen
# MIT License
"""
pyzil.zilliqa.node
~~~~~~~~~~~~

Zilliqa Node API.

:copyright: (c) 2019 by Gully Chen.
:license: MIT License, see LICENSE for more details.
"""

import socket
from jsonrpcclient.clients.socket_client import SocketClient


class Node:
    """Zilliqa Node API."""
    class NodeMethod:
        def __init__(self, node: "Node", method_name: str):
            self.node = node
            self.method_name = method_name

        def __call__(self, *params, **kwargs):
            resp = self.node.call(self.method_name, *params, **kwargs)
            return resp and resp.data and resp.data.result

    def __init__(self, host: str, port: int):
        self.host = host
        self.port = port

    def __str__(self):
        return "<Node {}:{}>".format(self.host, self.port)

    def __getattr__(self, item: str):
        return Node.NodeMethod(self, method_name=item)

    def call(self, method_name: str, *params, **kwargs):
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.connect((self.host, self.port))

        try:
            api_client = SocketClient(sock)
            return api_client.request(
                method_name, *params,
                trim_log_values=True, **kwargs
            )
        finally:
            sock.close()


LocalNode = Node("127.0.0.1", 4201)


if "__main__" == __name__:
    print(LocalNode.GetCurrentMiniEpoch())
    print(LocalNode.GetCurrentDSEpoch())
    print(LocalNode.GetNodeType())
    print(LocalNode.GetDSCommittee())
    print(LocalNode.GetNodeState())
    print(LocalNode.IsTxnInMemPool("txn_id"))
